const Controller = require('./controller');
const PubSub = require('pubsub-js');
const { menuEvent, requester } = require('../main/context');
const { expdata, grid, plane, peaks, matrix } = require('../server-state');

class PlaneController extends Controller {
    constructor(planePanel, gridPanel, expDataPanel, plane3D) {
        super();

        // Models definition.
        this._planePanel = planePanel;
        this._gridPanel = gridPanel;
        this._expDataPanel = expDataPanel;
        this._plane3D = plane3D;

        // Initialisation.
        this.init();

        // Actions handlers.
        PubSub.subscribe('update-plane-parameters', (msg, value) => { this.updatePlaneParameters(value); });
        PubSub.subscribe('update-grid', (msg, value) => { this.updateGrid(value); });
        PubSub.subscribe('update-expdata', (msg, value) => { this.updateExpData(value); });

        // Events handlers.
        PubSub.subscribe('change-peaks', () => { this.drawGrid(); });
        PubSub.subscribe('change-sample', () => { this.drawGrid(); });
        PubSub.subscribe('change-plane', () => { this.planeChange(); });
        PubSub.subscribe('change-grid', () => { this.gridChange(); });
        PubSub.subscribe('change-expdata', () => { this.expdataChange(); });

        menuEvent.on('change-mode', (event, value) => {
            this._plane3D.toggleGrid(value === 'snap');
        });
    }

    init() {
        this.drawGrid();
        // Set the mode to free.
        this._plane3D.toggleGrid(false);
        this._expDataPanel.setInputs(expdata);
        this._gridPanel.setInputs(grid);
        this._planePanel.setInputs(plane);
    }

    /**
     * Update experimental data display when modifications occur in the model.
     */
    expdataChange() {
        // Is it a possible scenario to need to refresh these data?
        this._expDataPanel.setInputs(expdata);
        this._plane3D.loadSqw(expdata);
    }

    /**
     * Update grid in response to grid change in model.
     */
    gridChange() {
        this._gridPanel.setInputs(grid);
        this.drawGrid();
    }

    /**
     * Update plane in response to a change in plane model.
     */
    planeChange() {
        this._planePanel.setInputs(plane);
        this.drawGrid();
    }

    /**
     * Update the grid 3D view.
     */
    drawGrid() {
        this._plane3D.drawGrid(peaks, plane, grid, expdata);
    }

    /**
     * Update plane parameters.
     */
    updatePlaneParameters(planeParam) {
        requester.request('setPlane', planeParam, (data) => {

            // Update local plane.
            plane.update(data.plane);

            // Update local matrix.
            matrix.update(data.matrix);

            // Update views.
            this._planePanel.setInputs(plane);
            this.drawGrid();

            PubSub.publish('change-plane', 0);
        });
    }

    /**
     * Update grid settings (delta, subdivision number, ...).
     */
    updateGrid(gridParam) {
        requester.request('setGrid', gridParam, () => {
            // Update local grid state.
            grid.update(gridParam);

            // Update views.
            this.drawGrid();
            this._gridPanel.setInputs(grid);

            PubSub.publish('change-grid', 0);
        });
    }

    /**
     * Update experimental data.
     */
    updateExpData(newExpData) {
        requester.request('setExpData', newExpData, () => {
            // Update the experiment data.
            expdata.update(newExpData);

            // Draw exp data on plane.
            this._plane3D.loadSqw(expdata);
        });
    }

}

module.exports = PlaneController;
