const View3D = require('./view3d');
const {vector3WorldToView} = require('../../utils/conversion');
const math = require('mathjs');
const xtal = require('../../../server/model/xtal/xtal');
const { peaks } = require('../../server-state');
const THREE = require('three');

// Value for determining if a peak is in the scattering plane.
const CMP_EPSILON = 1.0E-08;

class LatticeView3D extends View3D {
    constructor(scene, raycaster, renderer) {
        super(scene);

        this._renderer = renderer;

        this._sphereGroup = new THREE.Group();
        this._material = new THREE.MeshStandardMaterial({color: 0xffffff, transparent: true, opacity:peaks.sphereOpacity});
        this._toolTip = null;

        document.addEventListener('mousemove', (event) => {
            let intersects = raycaster.intersectObjects(this._sphereGroup.children);

            if(intersects.length > 0) {
                let intersect = intersects[0].object;
                this._toolTip = {
                    "name": intersect.name,
                    "structure_factor": intersect.structure_factor,
                    "intensity": intersect.peak_intensity,
                    "miller": intersect.miller,
                    "pos": [ intersect.position.x, intersect.position.y, intersect.position.z ],
                };
            } else {
                this._toolTip = null;
            }
            this._updateTooltip();
        });
    }

    /**
     * Update the tooltip display for lattice coordinate.
     */
    _updateTooltip() {
        if (this._toolTip === null) {
            $('#latticeCoord').hide();
            return;
        }

        const vecObj = new THREE.Vector3(this._toolTip.pos[0], this._toolTip.pos[1], this._toolTip.pos[2]);
        vecObj.project(this._scene.userData.camera);

        const xView = this._renderer.domElement.clientLeft + 0.5 * (1. + vecObj.x) * this._renderer.domElement.clientWidth;
        const yView = this._renderer.domElement.clientTop + 0.5 * (1. - vecObj.y) * this._renderer.domElement.clientHeight;

        let label = this._toolTip.name;

        // if the structure factor is available, display it
        if (this._toolTip.structure_factor !== null) {
            const [sf, sf2] = xtal.formatStructureFactor(this._toolTip.structure_factor);
            label += "<br>Structure factor: " + sf + "<br>Intensity: " + sf2;
        }

        $('#latticeCoord').css({
            top: yView, left:  xView
        }).html(label).show();
    }

    // If the number of peaks changes, redraw the lattice or if a or b vector changes, u or b matrix changes.
    redrawLattice(peaks, atoms, sample, plane, bMatrix, displayOnlyPeaksInPlane = false) {
        // Remove the sphere group from the scene before readding it.
        this._scene.remove(this._sphereGroup);

        // Remove all of the current spheres from the group backwards to avoid skipping spheres.
        for (let i = this._sphereGroup.children.length - 1; i >= 0; i--) {
            this._sphereGroup.remove(this._sphereGroup.children[i]);
        }

        // Draw the lattice while giving the upper and lower limits for the peaks in the xyz coordinates.
        this.drawLattice(peaks, atoms, sample, plane, bMatrix, displayOnlyPeaksInPlane);
    }

    // Draw the lattice.
    drawLattice(peaks, atoms, sample, plane, bMatrix, displayOnlyPeaksInPlane) {

        // Reset the material.
        this._material = new THREE.MeshStandardMaterial({color: 0xffffff, transparent: true, opacity:peaks.sphereOpacity});

        // Geometry and material of the peaks.
        let base_scale = 2.;
        let min_radius = 1;
        let max_radius = 8;
        let geometry = new THREE.SphereGeometry(base_scale, 16, 16);

        // get symmetric positions and scattering lengths, if possible
        let scattering_lengths = null;
        let sympos = null;
        if(atoms.atomnames !== null && atoms.atompos !== null
            && atoms.atomnames.length > 0) {
            scattering_lengths = xtal.getAllScatteringLengths(atoms.atomnames);
            sympos = xtal.getAllSymmetricPositions(sample.sg, atoms.atompos);
        }

        // Iterate on the peaks.
        for (let i = peaks.lowX; i <= peaks.upperX; i++) {
            for (let j = peaks.lowY; j <= peaks.upperY; j++) {
                for (let k = peaks.lowZ; k <= peaks.upperZ; k++) {

                    let lattice_vector = [i, j, k];

                    // is this peak forbidden by the space group?
                    if(xtal.isPeakAbsent(lattice_vector, sample.sg))
                        continue;

                    let cur_geometry = geometry;
                    let structure_factor = null;
                    let peak_intensity = null;

                    // calculate structure factor if we have all necessary data
                    if(scattering_lengths !== null && sympos !== null)
                    {
                        structure_factor = xtal.getStructureFactorDirect(lattice_vector,
                            scattering_lengths, sympos);
                        peak_intensity = xtal.getPeakIntensity(structure_factor);
                        if(math.equal(peak_intensity, 0))
                            continue;

                        let radius = 0.1 * base_scale * math.sqrt(peak_intensity);
                        if(radius > max_radius) radius = max_radius;
                        if(radius < min_radius) radius = min_radius;
                        cur_geometry = new THREE.SphereGeometry(radius, 16, 16);
                    }

                    // Vector for each sphere to be drawn.
                    let a = new THREE.Vector3(i, j, k);

                    // Create a new mesh.
                    let sphere = new THREE.Mesh(cur_geometry, this._material);

                    // Multiple the vector by the B matrix in order to get the coordinates for each sphere in R.
                    a.applyMatrix3(bMatrix);

                    // Name of each sphere is the original vector.
                    sphere.name = '(' + i + ',' + j + ',' + k + ')';
                    sphere.position.copy(vector3WorldToView(a));

                    // Used to determine if the sphere is in the scattering plane
                    sphere.miller = lattice_vector;
                    sphere.structure_factor = structure_factor;
                    sphere.peak_intensity = peak_intensity;

                    let inPlane = false;
                    let miller = new THREE.Vector3(sphere.miller[0], sphere.miller[1], sphere.miller[2]).applyMatrix3(bMatrix);

                    // Determine if the sphere is inside the plane.
                    let millVal = plane.abVector.x * miller.x +
                        plane.abVector.y * miller.y +
                        plane.abVector.z * miller.z;
                    if (millVal > -CMP_EPSILON && millVal < CMP_EPSILON) {

                        inPlane = true;

                        // To change the colour of a single object in the group, the material must be changed.
                        let inPlaneMaterial = new THREE.MeshStandardMaterial({color: 0xff0000, transparent: true, opacity:peaks.sphereOpacity});
                        sphere.material = inPlaneMaterial;
                    }

                    // (000)
                    if (i === 0 && j === 0 && k === 0) {
                        let yellowMaterial = new THREE.MeshStandardMaterial({color: 0xffff00, transparent: true, opacity:peaks.sphereOpacity});
                        sphere.material = yellowMaterial;
                    }

                    if (displayOnlyPeaksInPlane && inPlane || !displayOnlyPeaksInPlane) {
                        // Add sphere to group so they can all be added at one time.
                        this._sphereGroup.add(sphere);
                    }
                }
            }
        }

        // Add the sphere group to the scene.
        this._scene.add(this._sphereGroup);
    }
}

module.exports = LatticeView3D;