const View3D = require('./view3d');
const {vector3WorldToView, vector3ViewToWorld} = require('../../utils/conversion');
const { mapColor } = require('../../utils/map-color');
const THREE = require('three');
const { grid, matrix } = require('../../server-state');
const { requester } = require('../../main/context');

const X_DIR = 1;
const Y_DIR = 2;
const Z_DIR = 3;

const MAX_LOD = 7;
const MIN_LOD = 0;

class PlaneView3D extends View3D {
    constructor(scene, raycaster, mapControls) {
        super(scene);

        this._name = 'mainGrid';
        this._xAxis = new THREE.Vector3(1, 0, 0);
        this._yAxis = new THREE.Vector3(0, 1, 0);

        ///////// Grid mesh.
        this._material = new THREE.LineBasicMaterial({ vertexColors: THREE.VertexColors, linewidth: 1, depthTest: false });
        this._mesh = new THREE.LineSegments(new THREE.Geometry(), this._material);
        this._scene.add(this._mesh);

        ///////// Plane mesh.
        this._planeGeometry = new THREE.Geometry();
        this._planeGeometry.dynamic = true;
        this._planeGeometry.vertices = [
            new THREE.Vector3(0, 0, 0), new THREE.Vector3(0, 1, 0), new THREE.Vector3(1, 1, 0), new THREE.Vector3(1, 0, 0)
        ];

        // Faces.
        this._planeGeometry.faces.push(new THREE.Face3(0, 1, 2));
        this._planeGeometry.faces.push(new THREE.Face3(0, 2, 3));

        // UV coordinates (need to flip v coordinate!).
        this._planeGeometry.faceVertexUvs[0].push([new THREE.Vector2(0,1), new THREE.Vector2(0,0), new THREE.Vector2(1,0)]);
        this._planeGeometry.faceVertexUvs[0].push([new THREE.Vector2(0,1), new THREE.Vector2(1,0), new THREE.Vector2(1,1)]);

        // Plane material.
        let material = new THREE.MeshStandardMaterial({
            color: 0x666666, opacity: 0.75, transparent: true, side: THREE.DoubleSide
        });

        this._planeMesh = new THREE.Mesh(this._planeGeometry, material)
        this._planeMesh.renderOrder = 1;
        this._scene.add(this._planeMesh);

        // Register a sphere that can be snap on this grid.
        PubSub.subscribe('register-sphere', (message, value) => {
            if(value.gridName === this._name) {
                value.sphere.snapGrid = this;
            }
        });

        // Coordinate tooltip follow.
        document.addEventListener('mousemove', (event) => {
            let intersect = raycaster.intersectObjects([this._planeMesh]);

            if(intersect.length > 0) {
                const planePointInvA = vector3ViewToWorld(intersect[0].point);
                const planePointRLU = planePointInvA.clone().applyMatrix3(matrix.bMatrix_1);

                $('#coordinates').css({
                    top: event.layerY - 40, left:  event.layerX + 0
                }).html(
                    `RLU (${planePointRLU.x.toFixed(3)}, ${planePointRLU.y.toFixed(3)}, ${planePointRLU.z.toFixed(3)})</br>
                    \u212b\u207B\u00b9 (${planePointInvA.x.toFixed(3)}, ${planePointInvA.y.toFixed(3)}, ${planePointInvA.z.toFixed(3)})`
                ).show();
            } else {
                $('#coordinates').hide();
            }
        });

        this._currentLod = 0;

        mapControls.on('zoom', (event) => {

            let lod = -Math.log2(event.zoom);

            if (this._currentLod < MAX_LOD && lod - this._currentLod > 1) {
                // zoom in
                if (grid.adaptative) {
                    this._currentLod += lod - this._currentLod;
                    PubSub.publish('update-grid', { numDiv: 10 * (1 << this._currentLod), delta: grid.delta, adaptative: grid.adaptative });
                }
                
            } else if (this._currentLod > MIN_LOD && lod - this._currentLod < -1) {
                // zoom out
                if (grid.adaptative) {   
                    this._currentLod += lod - this._currentLod;             
                    PubSub.publish('update-grid', { numDiv: 10 * (1 << this._currentLod), delta: grid.delta, adaptative: grid.adaptative });
                }
            }
        });
    }

    /**
     * Project peaks on plane. 
     * This function define the x and y axes to draw the grid.
     */
    calculateAxes(plane) {
        let raycaster = new THREE.Raycaster();

        // Choose prefered direction to project peaks depending on plane inclination.
        this._direction = (Math.abs(plane.abVector.y) > 0.5) ? Y_DIR :
                          (Math.abs(plane.abVector.x) > 0.5) ? X_DIR : Z_DIR;

        // Set direction vector depending on prefered direction.
        let directionVector = (this._direction == Y_DIR) ? new THREE.Vector3(0, -1, 0) :
                              ((this._direction == X_DIR) ? new THREE.Vector3(-1, 0, 0) : new THREE.Vector3(0, 0, -1));
        directionVector.applyMatrix3(matrix.bMatrix);

        ////////// X AXIS PROJECTION /////////
        // Place the camera depending on prefered direction for the x axis.
        let position = (this._direction == Y_DIR) ? new THREE.Vector3(1, 1000, 0) :
                       ((this._direction == X_DIR) ? new THREE.Vector3(1000, 0, 1) : new THREE.Vector3(1, 0, 1000));
        position.applyMatrix3(matrix.bMatrix);

        raycaster.set(position, directionVector);
        raycaster.ray.intersectPlane(new THREE.Plane(plane.abVector), this._xAxis);
        this._xAxis = vector3WorldToView(this._xAxis);

        /////////////////////////
        /*let lineGeo = new THREE.Geometry();
        lineGeo.vertices.push(
            vector3WorldToView(position),
            vector3WorldToView(position.clone().add(directionVector.clone().multiplyScalar(1005)))
        );
        let line = new THREE.Line(lineGeo, new THREE.LineBasicMaterial({ 
            color: 0xffff00,
            linewidth: 5,
        }));
        line.name = "line1";
        this._scene.remove(this._scene.getObjectByName("line1"));
        this._scene.add( line );*/
        /////////////////////////

        ////////// Y AXIS PROJECTION /////////
        // Place the camera depending on prefered direction for the y axis.
        position = (this._direction == Y_DIR) ? new THREE.Vector3(0, 1000, 1) :
                   ((this._direction == X_DIR) ? new THREE.Vector3(1000, 1, 0) : new THREE.Vector3(0, 1, 1000));
        position.applyMatrix3(matrix.bMatrix);

        raycaster.set(position, directionVector);
        raycaster.ray.intersectPlane(new THREE.Plane(plane.abVector), this._yAxis);
        this._yAxis = vector3WorldToView(this._yAxis);

        /////////////////////////
        /*lineGeo = new THREE.Geometry();
        lineGeo.vertices.push(
            vector3WorldToView(position),
            vector3WorldToView(position.clone().add(directionVector.clone().multiplyScalar(1005)))
        );
        line = new THREE.Line(lineGeo, new THREE.LineBasicMaterial({ 
            color: 0xffff00,
            linewidth: 5,
        }));
        line.name = "line2";
        this._scene.remove(this._scene.getObjectByName("line2"));
        this._scene.add( line );*/
        /////////////////////////
    }

    /**
     * Calculate plane vertices coordinates.
     */
    calculatePlane(xMin, xMax, yMin, yMax) {
        this._plane00 = vector3ViewToWorld(this._xAxis.clone().multiplyScalar(xMin).add(this._yAxis.clone().multiplyScalar(yMin)));
        this._plane01 = vector3ViewToWorld(this._xAxis.clone().multiplyScalar(xMin).add(this._yAxis.clone().multiplyScalar(yMax)));
        this._plane11 = vector3ViewToWorld(this._xAxis.clone().multiplyScalar(xMax).add(this._yAxis.clone().multiplyScalar(yMax)));
        this._plane10 = vector3ViewToWorld(this._xAxis.clone().multiplyScalar(xMax).add(this._yAxis.clone().multiplyScalar(yMin)));
    }

    /**
     * Draw subdivisions on a given axis.
     */
    drawAxisSubdivisions(xAxis, yAxis, min, max, start, end, numDiv, delta, useSubdiv) {
        let vertices = [];
        let colors = [];
        let j = 0;

        // Color of the subdivision lines.
        let color = new THREE.Color(0x888888);
        // Color of the unit lines.
        let color2 = new THREE.Color(0x0000ff);

        // Unit lines loop.
        for(let i = start; i < end; i++) {
            let initValue = (useSubdiv) ? 0 : i;
            let stopCond = (useSubdiv) ? numDiv : i + 1;
            let incr = (useSubdiv) ? 1 : delta;
            let colorVal = (useSubdiv) ? 10 : 1;

            // Subdivision loop.
            for(let k = initValue; k <= stopCond; k += incr) {
                let delta = (useSubdiv) ? i + k / numDiv : k;

                let startPos = xAxis.clone().multiplyScalar(min).add(yAxis.clone().multiplyScalar(delta));
                let endPos = xAxis.clone().multiplyScalar(max).add(yAxis.clone().multiplyScalar(delta));

                vertices.push(startPos.x, startPos.y, startPos.z);
                vertices.push(endPos.x, endPos.y, endPos.z);

                // Choose the appropriate color.
                if(k % colorVal == 0) {
                    color2.toArray( colors, j); j += 3;
                    color2.toArray( colors, j); j += 3;
                } else {
                    color.toArray( colors, j); j += 3;
                    color.toArray( colors, j); j += 3;
                }
            }
        }

        return { vertices: vertices, colors: colors };
    }

    /*
     * Draw the scattering plane.
     */
    redrawScatteringPlane(expData) {
        // vertices
        this._planeGeometry.vertices[0].copy(vector3WorldToView(this._plane00));
        this._planeGeometry.vertices[1].copy(vector3WorldToView(this._plane01));
        this._planeGeometry.vertices[2].copy(vector3WorldToView(this._plane11));
        this._planeGeometry.vertices[3].copy(vector3WorldToView(this._plane10));
        this._planeGeometry.verticesNeedUpdate = true;
        this._planeGeometry.normalsNeedUpdate = true;
        this._planeGeometry.computeFaceNormals();
        this._planeGeometry.computeVertexNormals();
        this._planeGeometry.computeBoundingSphere();

        this.loadSqw(expData);
    }

    /**
     * Toggle grid visibility.
     */
    toggleGrid(value) {
        this._mesh.visible = value;
    }

    /**
     * Draw the grid.
     */
    drawGrid(peaks, plane, grid, expData) {

        this.calculateAxes(plane);
        this.calculatePlane(peaks.lowX, peaks.upperX, peaks.lowY, peaks.upperY);

        // Get vertices and colors arrays.
        // Always set usSubDiv to true because the delta value is not used.
        let hRes = this.drawAxisSubdivisions(this._xAxis, this._yAxis, peaks.lowX, peaks.upperX, peaks.lowY, peaks.upperY, grid.numDiv, grid.delta, true);
        let vRes = this.drawAxisSubdivisions(this._yAxis, this._xAxis, peaks.lowY, peaks.upperY, peaks.lowX, peaks.upperX, grid.numDiv, grid.delta, true);
        let vertices = vRes.vertices.concat(hRes.vertices);
        let colors = vRes.colors.concat(hRes.colors);
        
        let geometry = new THREE.BufferGeometry();
        geometry.addAttribute('position', new THREE.Float32BufferAttribute(vertices, 3));
        geometry.addAttribute('color', new THREE.Float32BufferAttribute(colors, 3));

        // Update mesh geometry with the new grid.
        this._mesh.geometry = geometry;

        this.redrawScatteringPlane(expData);
    }

    /*
     * Transform scattering plane texture UV coordinates to A^(-1) crystal coordinates
     */
    planeUV2InvAngs(u, v) {
        let vec00 = this._plane00.clone();
        let vec01 = this._plane01.clone();
        let vec10 = this._plane10.clone();

        vec10.sub(vec00);
        vec01.sub(vec00);
        vec10.multiplyScalar(u);
        vec01.multiplyScalar(v);
        vec00.add(vec01);
        vec00.add(vec10);

        return vec00;
    }

    /*
     * Transform scattering plane texture UV coordinates to RLU crystal coordinates
     */
    planeUV2RLU(u, v) {
        const B_1 = matrix.bMatrix_1;
        return this.planeUV2InvAngs(u, v).applyMatrix3(B_1);
    }

    /*
     * Loads a S(q,w) texture onto the scattering plane.
     */
    loadSqw(expdataModel) {

        // If none is selected, we set the material back to a gray background.
        if (expdataModel.sqwtype === "sqw-none") {
            delete this._planeMesh.material.map;
            this._planeMesh.material.color = new THREE.Color(0x666666);
            this._planeMesh.material.metalness = 0.5;
            this._planeMesh.material.roughness = 0.5;
            this._planeMesh.material.needsUpdate = true;
            return;
        }

        const E = parseFloat(expdataModel.texE);
        const minS = parseFloat(expdataModel.texminsqw);
        const maxS = parseFloat(expdataModel.texmaxsqw);
        const texWidth = parseInt(expdataModel.texwidth);
        const texHeight = parseInt(expdataModel.texheight);

        // create texture map
        let textureCanvas = document.createElement("canvas");
        textureCanvas.width = texWidth;
        textureCanvas.height = texHeight;
        let textureContext = textureCanvas.getContext("2d");
        textureContext.fillStyle = "rgba(0xff, 0xff, 0xff, 0xff)";
        textureContext.fillRect(0, 0, texWidth, texHeight);

        // create an S(q,w) texture
        let textureData = textureContext.getImageData(0, 0, texWidth, texHeight);

        // fill texture, iterating over texels
        let arrayData = [];

        for (let pixY = 0; pixY < texHeight; ++pixY) {
            //console.log("Calculating texture: " + parseInt(100. * pixY / (texHeight-1)) + "%.");
            
            let lineData = [];
            
            for (let pixX = 0; pixX < texWidth; ++pixX) {

                const coordU = pixX / (texWidth - 1);
                const coordV = pixY / (texHeight - 1);
                const coordRLU = this.planeUV2RLU(coordU, coordV);

                lineData.push([coordRLU.x, coordRLU.y, coordRLU.z, E]);
            }

            arrayData.push(lineData);
        }

        requester.request('sqw', { texWidth, texHeight, data: arrayData }, (data) => {

            for (let pixY = 0; pixY < texHeight; ++pixY) {
                const rowOffs = pixY*texWidth*4;
                
                let lineResult = data[pixY];

                for (let pixX = 0; pixX < texWidth; ++pixX) {

                    const pixOffs = rowOffs + pixX*4;
                    const color = mapColor(lineResult[pixX], minS, maxS);
                    textureData.data[pixOffs + 0] = color.r;
                    textureData.data[pixOffs + 1] = color.g;
                    textureData.data[pixOffs + 2] = color.b;
                    textureData.data[pixOffs + 3] = 0xff;
                }
            }

            textureContext.putImageData(textureData, 0, 0);

            // set texture
            delete this._planeMesh.material.color;
            this._planeMesh.material.map = new THREE.CanvasTexture(textureCanvas);
            this._planeMesh.material.metalness = 0;
            this._planeMesh.material.roughness = 1;
            this._planeMesh.material.needsUpdate = true;
        });
    }

    /**
     * Snap the position to the nearest point on the grid.
     */
    snap(position) {
        
        // Get the coord point in R
        let pos = vector3ViewToWorld(position.clone().applyMatrix3(matrix.bMatrix_1));

        let x = (this._direction == Y_DIR) ? pos.x : ((this._direction == X_DIR) ? pos.z : pos.x);
        let y = (this._direction == Y_DIR) ? pos.z : ((this._direction == X_DIR) ? pos.y : pos.y);

        let numDiv = grid.numDiv;
        let coord = new THREE.Vector2(Math.round(x * numDiv) / numDiv, Math.round(y * numDiv) / numDiv);
        /*let coord = null;
        if(this._gridModel.useSubdiv) {
            coord = new THREE.Vector2(Math.round(x * this._divisions) / this._divisions, Math.round(y * this._divisions) / this._divisions);
        } else {
            coord = new THREE.Vector2(Math.round(x / this._delta) * this._delta, Math.round(y / this._delta) * this._delta);
        }*/
        
        // Place the point correctly on the scattering plane
        let newPosition = new THREE.Vector3(
            this._xAxis.x * coord.x + this._yAxis.x * coord.y,
            this._xAxis.y * coord.x + this._yAxis.y * coord.y,
            this._xAxis.z * coord.x + this._yAxis.z * coord.y
        );

        return newPosition;
    }
}

module.exports = PlaneView3D;